import asyncio

from .methods import Methods
from .processor import Processor
from .serializer import Serializer


class AsyncSocketClient(Serializer, Processor, Methods):
    SIGNAL_ACT = -100
    SIGNAL_PING = -200
    ACT_TIMEOUT = 15
    PING_INTERVAL = 30

    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter
    is_connected: bool

    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self.reader = reader
        self.writer = writer
        super().__init__()


from abc import ABC
from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from .client import AsyncSocketClient

SIMPLE_TYPES = (int, float, str, bool)

CUSTOM_TYPE_HEADER = '__customtype__'
TUPLE_TYPE_NAME = 'builtins.tuple'
SET_TYPE_NAME = 'builtins.set'


class SerializerType(ABC):
    """ Class for create a serializer of a custom object type """

    INSTANCE: object
    default: "Serializer"

    @classmethod
    def serialize(cls, data: any) -> any:
        raise NotImplementedError

    @classmethod
    def deserialize(cls, data: any) -> any:
        raise NotImplementedError


class BytesSerializer(SerializerType):
    """ Default Bytes Serializer """

    INSTANCE = bytes

    @classmethod
    def serialize(cls, data: bytes) -> str:
        return data.hex()

    @classmethod
    def deserialize(cls, data: str) -> bytes:
        return bytes.fromhex(data)


class ByteArraySerializer(SerializerType):
    """ Default Bytearray Serializer """

    INSTANCE = bytearray

    @classmethod
    def serialize(cls, data: bytearray) -> str:
        return data.hex()

    @classmethod
    def deserialize(cls, data: str) -> bytearray:
        return bytearray.fromhex(data)


class Serializer(ABC):
    """ Serializer for simple data types """

    def __init__(self: "AsyncSocketClient"):
        super().__init__()
        self.add_serializer(BytesSerializer) # default bytes serializer
        self.add_serializer(ByteArraySerializer) # default bytearray serializer

    @staticmethod
    def get_data_type_str(data: Type[any]) -> str:
        return '{}.{}'.format(data.__module__, data.__name__)

    def serialize(self: "AsyncSocketClient", data: any) -> any:
        if data is None:
            return data

        elif isinstance(data, SIMPLE_TYPES):
            return data

        elif isinstance(data, list):
            return [self.serialize(value) for value in data]

        elif isinstance(data, tuple):
            return CUSTOM_TYPE_HEADER, TUPLE_TYPE_NAME, [self.serialize(value) for value in data]

        elif isinstance(data, set):
            return CUSTOM_TYPE_HEADER, SET_TYPE_NAME, [self.serialize(value) for value in data]

        elif isinstance(data, dict):
            return {name: self.serialize(value) for name, value in data.items()}

        else:
            data_type = self.get_data_type_str(type(data))
            serializer = self.custom_serializers.get(data_type, None)

            if not serializer:
                raise ValueError(f'unsupported data type: {data_type}')

            serialized_object = serializer.serialize(data)

            return CUSTOM_TYPE_HEADER, data_type, serialized_object

    def deserialize(self: "AsyncSocketClient", data: any) -> any:
        if data is None:
            return data

        elif isinstance(data, SIMPLE_TYPES):
            return data

        elif isinstance(data, list):

            if len(data) == 3 and data[0] == CUSTOM_TYPE_HEADER:
                if data[1] == TUPLE_TYPE_NAME:
                    return tuple([self.deserialize(value) for value in data[2]])
                elif data[1] == SET_TYPE_NAME:
                    return set([self.deserialize(value) for value in data[2]])
                elif data[1] in self.custom_serializers:
                    deserialized_object = self.custom_serializers[data[1]].deserialize(data[2])
                    return deserialized_object
                else:
                    raise ValueError(f'unsupported data type: {data[1]}')

            return [self.deserialize(value) for value in data]

        elif isinstance(data, dict):
            return {name: self.deserialize(value) for name, value in data.items()}

        else:
            raise ValueError(f'unsupported data type: {type(data)}')



import asyncio
import base64
import json
import struct
from abc import ABC
from typing import TYPE_CHECKING, Type, Dict

from .send_packet import SendPacket
from .serializer import SerializerType

if TYPE_CHECKING:
    from .client import AsyncSocketClient


class Methods(ABC):
    custom_serializers: Dict[str, Type[SerializerType]]

    def __init__(self):
        super().__init__()
        self.custom_serializers = {}

    @classmethod
    async def connect(cls: Type["AsyncSocketClient"], host: str, port: int):
        """ Connect to a server and Create socket client instance """

        reader, writer = await asyncio.open_connection(host, port)
        session = cls(reader, writer)
        await session._keep_alive()
        return session

    @property
    def is_connected(self: "AsyncSocketClient") -> bool:
        return self._keep_alive_task and not self._keep_alive_task.done() and not self.reader.at_eof()

    def add_serializer(self: "AsyncSocketClient", serializer: Type[SerializerType]):
        if not issubclass(serializer, SerializerType):
            raise ValueError('Unsupported serializer type')
        data_type = self.get_data_type_str(serializer.INSTANCE)
        serializer.default = self
        self.custom_serializers[data_type] = serializer

    async def send(self: "AsyncSocketClient", **data):
        """ Serialize and put Packet to send queue """

        # serialize
        data = self.serialize(data)

        # convert to JSON object
        json_data = json.dumps(data)

        # convert to bytes
        encoded_json_data = base64.b64encode(json_data.encode('utf-8'))

        # Send JSON response
        packet_length = struct.pack("!i", len(encoded_json_data))

        # put to queue
        packet = SendPacket(packet_length + encoded_json_data)
        await self._queue_send.put(packet)

        # waiting for execute
        await packet.sent.wait()

    async def receive(self: "AsyncSocketClient"):
        """ Waiting for income Packet """

        packet_data = await self._received_packets.get()

        # read JSON from bytes
        json_string = base64.b64decode(packet_data)
        data = json.loads(json_string)

        # deserialize
        return self.deserialize(data)

    async def close(self: "AsyncSocketClient"):
        self.writer.close()
        try:
            await asyncio.wait_for(self.writer.wait_closed(), timeout=10)
        except asyncio.TimeoutError:
            pass
        self._keep_alive_task.cancel()



import asyncio
import struct
import time
from abc import ABC
from typing import TYPE_CHECKING, Optional

from .send_packet import SendPacket

if TYPE_CHECKING:
    from .client import AsyncSocketClient


class Processor(ABC):
    _queue_send: asyncio.Queue["SendPacket"] # queue of send tasks
    _pending_packets: asyncio.Queue["SendPacket"] # queue of sent packets which waiting for acknowledging signals
    _received_packets: asyncio.Queue[bytes] # queue of received packets
    _keep_alive_task: Optional[asyncio.Task] # keep alive control and queue executor
    _last_ping_timestamp: float

    def __init__(self: "AsyncSocketClient"):
        super().__init__()
        self._queue_send = asyncio.Queue()
        self._pending_packets = asyncio.Queue()
        self._received_packets = asyncio.Queue()
        self._keep_alive_task = None
        self._last_ping_timestamp = time.time()

    async def _recv(self: "AsyncSocketClient", length: int) -> bytes:
        data = b''
        while left_bytes := length - len(data):
            data += await self.reader.read(left_bytes)
            await asyncio.sleep(.1)
        return data

    async def _send_ack(self: "AsyncSocketClient"):
        """
        Send Acknowledging signal.
        Used to send confirmation that the packet was successfully received.
        """
        signal = struct.pack("!i", self.SIGNAL_ACT)
        self.writer.write(signal)
        await self.writer.drain()

    async def _ping(self: "AsyncSocketClient"):
        """
        Send ping signal.
        And waiting for Acknowledging signal.
        """
        signal = struct.pack("!i", self.SIGNAL_PING)
        self.writer.write(signal)
        await self.writer.drain()

        ping_packet = SendPacket(self.SIGNAL_PING)
        await self._pending_packets.put(ping_packet)
        try:
            await asyncio.wait_for(ping_packet.sent.wait(), timeout=self.ACT_TIMEOUT)
        except asyncio.TimeoutError:
            raise ConnectionAbortedError(f'interrupt connection: client is not active')

    async def _keep_alive(self: "AsyncSocketClient"):
        async def _receiver_task():
            while self.is_connected:

                # check income packets
                try:
                    # receive packet header
                    packet_length = await asyncio.wait_for(self._recv(4), timeout=1)
                    packet_length = struct.unpack("!i", packet_length)[0]

                    # it is signal
                    if packet_length < 0:

                        # ping
                        if packet_length == self.SIGNAL_PING:
                            self._last_ping_timestamp = time.time()
                            await self._send_ack()

                        # acknowledging
                        elif packet_length == self.SIGNAL_ACT:
                            packet = await self._pending_packets.get()
                            packet.sent.set()
                            self._last_ping_timestamp = time.time()

                        else:
                            raise RuntimeError(f'got unrecognized signal: {packet_length}')

                    else:
                        # receive a full packet
                        packet_data = await self._recv(packet_length)

                        # send act signal
                        await self._send_ack()

                        # put to received packets queue
                        await self._received_packets.put(packet_data)
                except asyncio.TimeoutError:
                    pass

        async def _sender_task():
            ping_task = None
            while self.is_connected:
                try:
                    await asyncio.sleep(.01)

                    # time to ping
                    if time.time() > self._last_ping_timestamp + self.PING_INTERVAL:
                        if not ping_task:
                            ping_task = asyncio.create_task(self._ping())
                        elif ping_task.done():
                            if ping_task.exception():
                                raise ping_task.exception()
                            ping_task = None
                            self._last_ping_timestamp = time.time()

                    # execute send
                    if not self._queue_send.empty():
                        send_packet = await self._queue_send.get()

                        # send
                        self.writer.write(send_packet.data)
                        await self.writer.drain()

                        # waiting for act signal
                        await self._pending_packets.put(send_packet)

                except asyncio.TimeoutError:
                    continue

        async def _keep_alive_task():
            await asyncio.gather(_receiver_task(), _sender_task())

        self._keep_alive_task = asyncio.create_task(_keep_alive_task())


import asyncio
from abc import ABC
from contextlib import closing
from typing import TYPE_CHECKING

from .. import AsyncSocketClient

if TYPE_CHECKING:
    from .server import AsyncSocketServer


class Processor(ABC):

    async def _connection_handler_wrapper(self: "AsyncSocketServer", reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        """ Safe Wrapper for income connections """
        async def session_wrapper():
            with closing(writer):
                session = AsyncSocketClient(reader, writer)
                await session._keep_alive()
                return await self.connection_handler(session)
        self.sessions.append(asyncio.create_task(session_wrapper()))

    async def __aenter__(self: "AsyncSocketServer"):
        return self

    async def __aexit__(self: "AsyncSocketServer", exc_type, exc_value, traceback):
        await self.close()


import asyncio
from typing import List, Callable

from .processor import Processor


class AsyncSocketServer(Processor):
    port: int
    server: asyncio.Server
    sessions: List[asyncio.Task]
    connection_handler: Callable

    @classmethod
    async def start(cls, ports_range: List[int], connection_handler: Callable):
        """ Start a server with given ports range """

        # constructor
        server_wrap = cls()
        server_wrap.sessions = []
        server_wrap.connection_handler = connection_handler

        # find free port in range
        for port in ports_range:
            try:

                # run server
                protocol_server = await asyncio.start_server(server_wrap._connection_handler_wrapper, '0.0.0.0', port)
                server_wrap.port = port
                server_wrap.server = protocol_server

                return server_wrap

            except OSError:
                pass
        else:
            raise RuntimeError(f'all ports in range {ports_range} are busy')

    async def idle(self):
        """ Keep the Server active """
        await self.server.serve_forever()

    async def close(self):
        """ Close the server """

        try:
            # close server
            self.server.close()

            # waiting for closing
            await asyncio.wait_for(self.server.wait_closed(), timeout=10)
        except asyncio.TimeoutError:
            pass

        # force cancel of not closed tasks
        [task.done() or task.cancel() for task in self.sessions]


from aiosocketproto.client import AsyncSocketClient
from aiosocketproto.client.serializer import SerializerType
from aiosocketproto.server import AsyncSocketServer

connect = AsyncSocketClient.connect
start_server = AsyncSocketServer.start

__version__ = '0.0.14'



import setuptools

import aiosocketproto

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiosocketproto",
    version=aiosocketproto.__version__,
    author="Roman D.",
    description="A simple and universal socket protocol written in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/webdenisenko/aiosocketproto",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10'
)
